﻿/*
 * C'Gol
 * Project based on Conway's Game of Life
 * 
 * Made by:
 * Alexander Noss
 * Christian Rautenberg
 * David Zentner
 * Philipp Erler
 * Stephan Schneider
 * 
 * As IT project for summer-semester 2011 (Group IT2D)
 * 
 * ####################################################################
 * MainForm
 * Part of C'Gol GUI
 * ####################################################################
 * 
 * Last change:     2011 | 07 | 04
 * 
 * Mail me for feedback, version requests etc.: 	ph.erler@gmx.net
 * Look at http://www.hs-esslingen.de/~pherit00/website/index.htm
 * .Net - Framework 4.0 must be installed.
 * 
 */


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Runtime.InteropServices;

namespace CGol
{
    public partial class MainForm : Form
    {
        #region dllImports
        //get field size
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern int getFieldSize();

        //set field size
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern void setFieldSize(int size);

        //get field
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern int getCell(int x, int y);

        //update field
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern int updateField();

        //create new field
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern void createArray(int size);

        //create new field and fill randomly
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern void randomField();

        //load field
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern int loadField(byte fileNum);

        //save field
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern int saveField(byte fileNum);

        //switch cell
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern void switchCell(int x, int y);
        
        //delete field
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern void deleteField();
        
	    //init dll
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern void initDll();
        
	    // get number of living cells
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
	    static extern int getAlive();

	    // get number of dead cells
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
	    static extern int getDead();

	    // get number of born cells
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
	    static extern int getBorn();

	    // get number of gone cells
        [DllImport("cgolc.dll", CallingConvention = CallingConvention.Cdecl)]
        static extern int getGone();

        #endregion

        AboutForm frmAbout;                                             //About form
        frmFiles frmFile;                                               //File operation form
        frmCreate frmSize;                                              //set Field size form
        int fieldSize;                                                  //size of field
        float cellSize;                                                 //size of 1 cell
        private Pen rasterPen;                                          //pen for raster
        private Brush cellPen;                                          //brush for cell
        bool[,] field;                                                  //bool field
        int cycles;                                                     //number of cycles

        //stats
        int[] alive;                                                    //living cells
        int[] dead;                                                     //dead cells
        int[] born;                                                     //born cells
        int[] gone;                                                     //gone cells
        Pen alivePen;                                                   //pen for living
        Pen deadPen;                                                    //pen for dead
        Pen bornPen;                                                    //pen for born
        Pen gonePen;                                                    //pen for gone

        //constructor
        public MainForm()
        {
            InitializeComponent();                                      //controls

            try
            {
                initDll();
            }
            catch (Exception)
            {
                MessageBox.Show("An Error occurred while initializing the DLL. Make sure 'cgolc.dll' is in the same folder as this application.");
                Application.Exit();
            }

            //init field and draw
            updateCellSize();                                           //calc cell size
            drawCurrentField();                                         //update field

            //init cycles
            resetCycleAndGraph();
            printCycles();

            //set drawing mode and pens
            pictureBox.CreateGraphics().SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighSpeed;
            rasterPen = new Pen(System.Drawing.Color.LightGray);        //pens for raster
            cellPen = new SolidBrush(System.Drawing.Color.GreenYellow); //and living cells
            alivePen = new Pen(System.Drawing.Color.GreenYellow);       //pen for living
            deadPen = new Pen(System.Drawing.Color.Black);              //pen for dead
            bornPen = new Pen(System.Drawing.Color.Blue);               //pen for born
            gonePen = new Pen(System.Drawing.Color.Red);                //pen for gone

            //load startup-field if existing
            if (System.IO.File.Exists("0.txt"))
                loadFieldFromFile(System.Convert.ToByte('0'));
        }

        #region Controls

        //close application
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        //clean up field when closing app
        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            //try ... catch because this is executed when the app closes due to missing dll
            try
            {
                deleteField();
            }
            catch (Exception)
            {
            }
        }

        //show about form
        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            frmAbout = new AboutForm();
            frmAbout.Show();
        }

        //change update frequency
        private void trackBar_Scroll(object sender, EventArgs e)
        {
            lblFrequency.Text = String.Format("Update every {0} ms", trackBar.Value);
            timer.Interval = trackBar.Value;
        }

        //paint picturebox
        private void pictureBox_Paint(object sender, PaintEventArgs e)
        {
            Draw_All(e.Graphics);
            e.Graphics.Flush(System.Drawing.Drawing2D.FlushIntention.Sync);
        }

        //update field and show current
        private void timer_Tick(object sender, EventArgs e)
        {
            //update cycle count and show it
            increaseCycles();
            printCycles();

            //update field and show it
            updateCellSize();
            updateField();
            drawCurrentField();
        }

        //redraw stat graph
        private void picBoxStats_Paint(object sender, PaintEventArgs e)
        {
            //clear
            e.Graphics.Clear(System.Drawing.Color.Black);

            //only if valid field exists
            if (alive == null || fieldSize < 1)
                return;

            //move points to right
            for (int n = picBoxStats.Width - 1; n > 0; n--)
            {
                alive[n] = alive[n - 1];
                dead[n] = dead[n - 1];
                born[n] = born[n - 1];
                gone[n] = gone[n - 1];
            }

            //add new point
            alive[0] = getAlive();
            dead[0] = getDead();
            born[0] = getBorn();
            gone[0] = getGone();

            //draw lines / rects
            for (int n = 0; n < picBoxStats.Width; n++)
            {
                //alive
                if (alive[n] > 0)
                    e.Graphics.DrawLine(alivePen, n, picBoxStats.Height - (int)Math.Round(((double)picBoxStats.Height * 
                        (double)alive[n] / (double)(fieldSize * fieldSize)), 0), n, picBoxStats.Height);
                //born
                if (born[n] > 0)
                    e.Graphics.DrawRectangle(bornPen, n, picBoxStats.Height - (int)Math.Round(((double)picBoxStats.Height * 
                        (double)born[n] / (double)(fieldSize * fieldSize)), 0), 1, 1);
                //gone
                if (gone[n] > 0)
                    e.Graphics.DrawRectangle(gonePen, n, picBoxStats.Height - (int)Math.Round(((double)picBoxStats.Height * 
                        (double)gone[n] / (double)(fieldSize * fieldSize)), 0), 1, 1);
            }
        }

        //show readme
        private void readMeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                System.Diagnostics.Process.Start(Application.StartupPath + @"\readme.txt");
            }
            catch (System.ComponentModel.Win32Exception ex)
            {
                MessageBox.Show("Error while openening readme file: " + ex.Message);
            }
            catch (System.IO.FileNotFoundException ex)
            {
                MessageBox.Show("File not found error: " + ex.Message);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Unknown error: " + ex.Message);
            }
        }

        //create new field and draw it
        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int oldSize = getFieldSize();
            if (oldSize == 0)
                oldSize = 20;

            frmSize = new frmCreate(oldSize);
            frmSize.setSizeFunction(new frmCreate.setSize(setSize));
            frmSize.Show();
        }

        //create randomly filled field and draw it
        private void fillrandomToolStripMenuItem_Click(object sender, EventArgs e)
        {
            randomField();
            drawCurrentField();
        }

        //load field
        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /*
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                string path = openFileDialog.FileName;
                
                loadField(path.ToCharArray());
                getField();
            }*/

            frmFile = new frmFiles(frmFiles.LoadSaveState.Load);
            frmFile.setFileFunction(new frmFiles.fileFunction(loadFieldFromFile));
            frmFile.Show();
        }

        //save field
        private void saveFieldToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /*
            SaveFileDialog oFDia = new SaveFileDialog();

            oFDia.Filter = "C'Gol Fields|*.cgol|Textfiles|*.txt|All files|*.*";
            oFDia.FilterIndex = 1;

            if (oFDia.ShowDialog() == DialogResult.OK)
            {
                string path = oFDia.FileName;

                saveField(path.ToCharArray());
            }*/

            frmFile = new frmFiles(frmFiles.LoadSaveState.Save);
            frmFile.setFileFunction(new frmFiles.fileFunction(saveFieldToFile));
            frmFile.Show();
        }

        //click on picturebox to switch content of a cell
        private void pictureBox_MouseDown(object sender, MouseEventArgs e)
        {
            if (fieldSize < 1)
                return;

            //get cell under mouse
            Point p = new Point(-1, -1);

            try
            {
                p = getCellUnderPos(e.X, e.Y);
            }
            catch (Exception)
            {
            }

            //switch content of this cell
            if (p.X != -1 && p.X < fieldSize && p.Y < fieldSize)
                switchCell(p.X, p.Y);

            //show new field
            drawCurrentField();
        }

        //start and stop timer, show state
        private void butTmr_Click(object sender, EventArgs e)
        {
            if (!timer.Enabled)
            {
                timer.Enabled = true;
                butTmr.Text = "Stop";
            }
            else
            {
                timer.Enabled = false;
                butTmr.Text = "Start";
            }
        }

        #endregion

        //get current field
        private void getField()
        {
            if (fieldSize < 1)
                return;

            //create new field
            field = new bool[fieldSize, fieldSize];

            //get all cells
            for (int x = 0; x < fieldSize; x++)
            {
                for (int y = 0; y < fieldSize; y++)
                {
                    field[x, y] = System.Convert.ToBoolean(getCell(x, y));
                }
            }
        }

        //calc size of cell
        private void updateCellSize()
        {
            fieldSize = getFieldSize();

            if (fieldSize > 0)
                cellSize = ((float)pictureBox.Width / (float)fieldSize);
        }

        /// <summary>
        /// Return cell coords under mouse.
        /// Mousepos must be coords on picturebox.
        /// </summary>
        /// <param name="mouseX"></param>
        /// <param name="mouseY"></param>
        /// <returns></returns>
        private Point getCellUnderPos(int mouseX, int mouseY)
        {
            int x;
            int y;

            if (cellSize == 0)
            {
                updateCellSize();
            }
            
            x = System.Convert.ToInt32(Math.Floor((float)mouseX / cellSize));
            y = System.Convert.ToInt32(Math.Floor((float)mouseY / cellSize));

            return new Point(x, y);
        }

        #region Drawing

        //draw whole field
        private void Draw_All(Graphics graph)
        {
            Fill_BG(graph);                                                                     //clear
            Draw_Raster(graph);                                                                 //draw raster
            Draw_Cells(graph);                                                                  //draw cells
        }

        //show current field
        private void drawCurrentField()
        {
            getField();
            pictureBox_Paint(this, new PaintEventArgs(pictureBox.CreateGraphics(), 
                new Rectangle(0, 0, pictureBox.Width, pictureBox.Height)));
            updateStats();
            drawStats();
        }

        //clear picturebox with black
        private void Fill_BG(Graphics Graphic)
        {
            Graphic.Clear(System.Drawing.Color.Black);
        }

        //draw raster
        private void Draw_Raster(Graphics Graphic)
        {
            for (int i = 0; i < fieldSize; i++)
            {
                try
                {
                    int pos = System.Convert.ToInt32(i * cellSize);                             //for both x and y

                    //y-direction raster lines
                    Graphic.DrawLine(rasterPen, pos, 0, pos, pictureBox.Height);

                    //x-direction raster lines
                    Graphic.DrawLine(rasterPen, 0, pos, pictureBox.Width, pos);
                }
                catch (Exception)
                { return; }
            }
        }

        //draw living cells
        private void Draw_Cells(Graphics Graphic)
        {
            if (field == null)                                                                  //skip if no field
                return;

            for (int y = 0; y < fieldSize; y++)
            {
                for (int x = 0; x < fieldSize; x++)
                {
                    if (field[y, x] == true)
                    {
                        Graphic.FillEllipse(cellPen, x * cellSize, y * cellSize, cellSize, cellSize);
                    }
                }
            }
        }

        //update status bar with stat labels
        private void updateStats()
        {
            //get number of neccessary dezimals
            string dezimals = "";
            for (int n = 0; n < Math.Ceiling(Math.Log10(fieldSize * fieldSize)); n++)
            {
                dezimals += "0";
            }
            if (dezimals == "")
                dezimals = "0";

            toolStripStatAlive.Text = (getAlive()).ToString("Alive: " + dezimals);
            toolStripStatDead.Text = (getDead()).ToString("Dead: " + dezimals);
            toolStripStatBorn.Text = (getBorn()).ToString("Born: " + dezimals);
            toolStripStatGone.Text = (getGone()).ToString("Gone: " + dezimals);
        }

        //redraw stat graph
        private void drawStats()
        {
            picBoxStats_Paint(this, new PaintEventArgs(picBoxStats.CreateGraphics(),
                new Rectangle(0, 0, picBoxStats.Width, picBoxStats.Height)));
        }

        #endregion

        #region DelegateFunctions

        //function used by delegate on frmCreate; create new field of given size
        public void setSize(int size)
        {
            createArray(size);
            updateCellSize();
            resetCycleAndGraph();
            printCycles();

            drawCurrentField();
        }

        //function used by delegate on frmFiles
        public void saveFieldToFile(byte fileNumber)
        {
            int res = saveField(fileNumber);
            if (res != 0)
                MessageBox.Show("Error while saving field");
        }

        //function used by delegate on frmFiles
        public void loadFieldFromFile(byte fileNumber)
        {
            int res = loadField(fileNumber);
            if (res != 0)
                MessageBox.Show("Error while loading field");

            resetCycleAndGraph();
            printCycles();
            updateCellSize();
            drawCurrentField();
        }

        #endregion

        #region CycleCounter

        //show number of cycles in form title bar
        private void printCycles()
        {
            this.Text = "C'Gol - Cycle " + cycles.ToString();
        }

        //increase cycles
        private void increaseCycles()
        {
            cycles++;
        }

        //reset cycles and stat graph
        public void resetCycleAndGraph()
        {
            cycles = 0;

            //create array
            alive = new int[picBoxStats.Width];                         //living cells
            dead = new int[picBoxStats.Width];                          //dead cells
            born = new int[picBoxStats.Width];                          //born cells
            gone = new int[picBoxStats.Width];                          //gone cells
            //initialize
            for (int n = 0; n < picBoxStats.Width; n++)
            {
                alive[n] = dead[n] = born[n] = gone[n] = -1;
            }
        }

        #endregion

    }
}
